/******************************************************************************
 *
 *	PalmKart mini
 *	ver 1.3.3
 *	2008.11 / 2009.5
 *	ChoiKyun
 *
 *****************************************************************************/
#include <PalmOS.h>
#include <PalmOSGlue.h>
#include <SonyCLIE.h>
#include "defrsc.h"

#define GLOBAL_VALUE_DEFINE
#include "game.h"



/******************************************************************************
 *	vg^Cv
 *****************************************************************************/
Boolean check_hard_key( EventPtr );
void get_pen_input( void );
void check_enter( EventPtr );
void init_main_form( FormPtr );
void make_bitmap_offscreen( void );
Err rom_check( UInt32, UInt16 );
void *get_object_ptr( UInt16 );
Boolean main_form_command( UInt16 );
Boolean main_form_handler( EventPtr );
Boolean app_handle_event( EventPtr );
void event_loop( void );
Err app_start( void );
void app_stop( void );
UInt32 my_app_main( UInt16, MemPtr, UInt16 );
void key_preset( void );
UInt32 set_event_interval( void );



/************************************************
 *	n[hL[
 ***********************************************/
void
get_hard_key( void )
{
  Key.keep = Key.single;
  Key.keep ^= 0xffffffff;

  Key.single = KeyCurrentState();
  Key.keep &= Key.single;
}



/********************************************************************
 *	X^[eBbNȊ֐Q
 *******************************************************************/


/************************************************
 *	L[vZbg
 ***********************************************/
static void
key_preset( void )
{
  if( FrmGlueNavIsSupported() )
    {
      // 5-way
      Keyset.left = keyBitRockerLeft;
      Keyset.right = keyBitRockerRight;
      Keyset.up = KEY_UP;
      Keyset.center = keyBitRockerCenter;
      Keyset.acc =   KEY_2 | KEY_4;
      Keyset.brake = KEY_1 | KEY_3;
    }
  else
    {
      // nomal
      Keyset.left = KEY_1;
      Keyset.right = KEY_2;
      Keyset.up = KEY_UP;
      Keyset.center = 0;
      Keyset.acc =   KEY_4;
      Keyset.brake = KEY_3;
    }
}



/************************************************
 *	\[Xrbg}bvItXN[쐬
 ***********************************************/
static void
make_bitmap_offscreen( void )
{
  int i;
  UInt16 error;
  Coord width, height;
  MemHandle h;
  BitmapPtr p;


  // LN^@ItXN[쐬
  for( i = 0; i <= MAX_BMP; i++ )
    {
      h = DmGetResource( bitmapRsc, BMP_BASE_ID + i );
      p = MemHandleLock( h );

      if( Screen.os5 )
        BmpGlueGetDimensions( p, &width, &height, NULL );
      else
        {
          width = p ->width;
          height = p ->height;
        }

      Bmp.handle[i] = WinCreateOffscreenWindow( width, height, screenFormat, &error );
      ErrFatalDisplayIf( error, "Error loading images" );

      WinSetDrawWindow( Bmp.handle[i] );
      WinDrawBitmap( p, 0, 0 );

      Bmp.ptr[i] = BmpGetBits( WinGetBitmap( Bmp.handle[i] ) );
      Bmp.extent[i].x = width;
      Bmp.extent[i].y = height;

      if( Screen.os5 )
      {
          BmpSetTransparentValue( WinGetBitmap( Bmp.handle[i] ), TRANS_COLOR ); // ߐFݒ
      }
    
      MemPtrUnlock( p );
      DmReleaseResource( h );
    }

  WinSetDrawWindow( 0 );
}



/************************************************
 *	n[hL[̃`FbN
 ***********************************************/
static Boolean
check_hard_key( EventPtr event_p )
{
  Boolean handled = false;


  switch( event_p ->eType )
    {
    case keyDownEvent:
      if( TxtGlueCharIsVirtual( event_p ->data.keyDown.modifiers,
                                event_p ->data.keyDown.chr) )
        {
          switch( event_p ->data.keyDown.chr )
            {
            case vchrHard1:
            case vchrHard2:
            case vchrHard3:
            case vchrHard4:
              handled = true;
              break;

            case vchrRockerRight:
            case vchrRockerLeft:
            case vchrRockerCenter:
            case vchrPageDown:
            case vchrPageUp:
              if( Game.active ) handled = true;
              break;
            }
        }
    }

  return handled;
}



/************************************************
 *	Cxg[ṽEFCgݒ
 ***********************************************/
static UInt32
set_event_interval( void )
{
  Int32 wait;


  wait = (Timer.interval + Timer.frame_tick) - TimGetTicks();
  if( wait < 0 ) wait = 0;

  return wait;
}



/************************************************
 *	ANeBu`FbN
 ***********************************************/
static void
check_enter( EventPtr event_p )
{
  switch( event_p ->eType )
    {
    case winEnterEvent:
      if( event_p ->data.winEnter.enterWindow == (WinHandle) FrmGetFormPtr( MainForm )
          && event_p ->data.winEnter.enterWindow == (WinHandle) FrmGetFirstForm() )
        Game.active = true;
      break;

    case winExitEvent:
      if( event_p ->data.winExit.exitWindow == (WinHandle) FrmGetFormPtr( MainForm ) )
        Game.active = false;
      break;
    }
}



/************************************************
 *	ROM̃o[W`FbN
 *	PalmOS 3.5ȍ~H
 ***********************************************/
static Err
rom_check(UInt32 requiredVersion, UInt16 launchFlags)
{
  UInt32 rom_ver;

  // See if we're on in minimum required version of the ROM or later.
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_ver);

  if( rom_ver < requiredVersion )
    {
      if( ( launchFlags & ( sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp ) )
          == ( sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp ) )
        {
          FrmAlert( RomIncompatibleAlert );

          // Palm OS 1.0 will continuously relaunch this app unless we switch to 
          // another safe one.
          if( rom_ver < ourMinVersion )
            {
              AppLaunchWithCommand( sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL );
            }
        }

      return sysErrRomIncompatible;
    }

  return errNone;
}



#if 0
/************************************************
 *	tH[̃IuWFNg̃|C^擾
 ***********************************************/
static void *
get_object_ptr( UInt16 object_ID )
{
  FormPtr frm_p;


  frm_p = FrmGetActiveForm();
  return FrmGetObjectPtr( frm_p, FrmGetObjectIndex( frm_p, object_ID ) );
}
#endif



/************************************************
 *	CtH[j[R}h
 ***********************************************/
static Boolean
main_form_command( UInt16 command )
{
  Boolean handled = false;
  FormPtr frm_p;


  switch( command )
    {
    case MainOptionsNewGame:
      MenuEraseStatus( 0 );

      // j[Q[
      clear_screen();
      show_title();
      Game.state = GAME_TITLE;
      Timer.interval = TimGetTicks();

      handled = true;
      break;

    case MainOptionsBestRecords:
      // XRALO
      show_ranking();

      Timer.interval = TimGetTicks();
      handled = true;
      break;

    case MainOptionsAboutPalmCartmini:
      // AoEgʂ̕\
      MenuEraseStatus( 0 );
      frm_p = FrmInitForm( AboutForm );
      FrmDoDialog( frm_p );
      FrmDeleteForm( frm_p );

      Timer.interval = TimGetTicks();
      handled = true;
      break;

    case MainOptionsHelp:
      // wv
      FrmHelp( key_helpString );
      Timer.interval = TimGetTicks();
      handled = true;
      break;
    }

  return handled;
}



/************************************************
 *	CtH[Cxgnh
 ***********************************************/
static Boolean
main_form_handler( EventPtr event_p )
{
  Boolean handled = false;
  FormPtr frm_p;


  switch( event_p ->eType )
    {
    case nilEvent:
      // Q[̃C
      play_game();
      break;

    case menuEvent:
      return main_form_command( event_p ->data.menu.itemID );

    case frmOpenEvent:
      frm_p = FrmGetActiveForm();
      FrmDrawForm( frm_p );
      handled = true;
      break;

    case penUpEvent:
      break;

    case penMoveEvent:
      break;

    case penDownEvent:
      if( Game.state == GAME_TITLE )
        {
          if( (UInt16) event_p ->screenX < 92 && (UInt16) event_p ->screenY < 80 )
            {
              Owata.on ^= 1;
              set_se( click );
              handled = true;
            }
          else if( (UInt16) event_p ->screenY < 80 )
            {
              if( ++Engine.cc > MODE_100CC ) Engine.cc = 0;
              change_engine();
              set_se( click );
              handled = true;
            }
          else if( (UInt16) event_p ->screenY > 150 )
            {
              Engine.cc = MODE_150CC;
              change_engine();
              set_se( click );
              handled = true;
            }
        }
      break;
    }

  return handled;
}



/************************************************
 *	tH[̃[h
 ***********************************************/
static Boolean
app_handle_event( EventPtr event_p )
{
  UInt16 form_id;
  FormPtr frm_p;


  if( event_p ->eType == frmLoadEvent )
    {
      form_id = event_p ->data.frmLoad.formID;
      frm_p = FrmInitForm( form_id );
      FrmSetActiveForm( frm_p );

      switch( form_id )
        {
        case MainForm:
          FrmSetEventHandler( frm_p, main_form_handler );
          break;
        }

      return true;
    }

  return false;
}



/************************************************
 *	Cxg[v
 ***********************************************/
static void
event_loop( void )
{
  UInt16 error;
  EventType event;


  do
    {
      // Cxg擾
      EvtGetEvent( &event, 0 );//set_event_interval() );

      // ANeBu
      check_enter( &event );

      // n[hL[̓XLbv
      if( !check_hard_key( &event ) )
        if( SysHandleEvent( &event ) ) continue;

      if( !MenuHandleEvent( 0, &event, &error ) )
        if( !app_handle_event( &event ) )
          FrmDispatchEvent( &event );
    }
  while( event.eType != appStopEvent );
}



/************************************************
 *	AvJn
 ***********************************************/
static Err
app_start( void )
{
  Err err = errNone;
  GamePrefType pref;
  UInt16 size;
  Int16 ver;
  int i;


  // XN[
  err = init_screen();
  if( err )
    {
      FrmAlert( NonSupportedScreenAlert );
      return err;
    }

  // Wϊe[u
  err = open_geometry_db();
  if( err )
    {
      FrmAlert( GeometryDBErrorAlert );
      return err;
    }

  // S[Xgf[^x[X
  err = open_ghost_db();
  if( err )
    {
      FrmAlert( GhostDBErrorAlert );
      return err;
    }

  // R[X}bvf[^x[X
  err = open_coursemap_db();
  if( err )
    {
      FrmAlert( CourseMapDBErrorAlert );
      return err;
    }

  make_bitmap_offscreen();

  // [Nm
  // R[Xobt@
  Course.field_p = (UInt8 *)MemGluePtrNew( sizeof(UInt8) * COURSE_WIDTH * COURSE_HEIGHT );
  ErrFatalDisplayIf( !Course.field_p, "Error allocating Buffers" );
  // S[Xgobt@
  Ghost.work_p = (GhostRecType *)MemPtrNew( REC_GHOST_SIZE );
  ErrFatalDisplayIf( !Ghost.work_p, "Error allocating Buffers" );

  SysRandom ( TimGetTicks() );
  AutoOff.seconds = SysSetAutoOffTime( AUTO_OFF_TIME );

  Sound.on = BGM_ON_SE_ON;
  Sound.volume = PrefGetPreference( prefGameSoundVolume );

  // nCXRA
  init_hiscore();
  // R[X
  Course.no = 0;
  // [h
  Owata.on = false;
  init_engine();

  // vt@X擾
  size = sizeof( GamePrefType );
  ver = PrefGetAppPreferences( appFileCreator, appPrefID, &pref, &size, true );
  if( ver == noPreferenceFound
      || ver != appPrefVersionNum
      || size != sizeof( GamePrefType) )
    {
      // 
      init_hiscore();
    }
  else
    {
      // vt@X̃Rs[
      MemMove( &Rank, &pref, sizeof( RankingType ) );
      Sound.on = pref.sound;
      Owata.on = pref.owata_on;
      Engine.cc = pref.engine_cc;
    }

  // S[XgItO
  for( i = 0; i < COURSE_MAX; i++ )
    Ghost.on[i] = true;

  init_course();
  key_preset();
  init_bgm();
  change_engine();

  show_title();

  Game.state = GAME_TITLE;
  Timer.interval = TimGetTicks();
//  Timer.frame_tick = DEF_EVENT_TICK;
  Timer.frame_tick = SysTicksPerSecond() / SEC;

  return err;
}



/************************************************
 *	AvI
 ***********************************************/
static void
app_stop( void )
{
  int i;
  UInt16 hr_ref_num;
  GamePrefType pref;


  // vt@X̃Rs[
  MemMove( &pref, &Rank, sizeof( RankingType ) );
  pref.sound = Sound.on;
  pref.owata_on = Owata.on;
  pref.engine_cc = Engine.cc;
  PrefSetAppPreferences( appFileCreator, appPrefID, appPrefVersionNum,
                         &pref, sizeof( GamePrefType ), true );

  for( i = 0; i <= MAX_BMP; i++ )
    WinDeleteWindow( Bmp.handle[i], false );

  // obNXN[
  WinDeleteWindow( Screen.back_h, false );

  // [N
  MemPtrFree( Course.field_p );
  MemPtrFree( Ghost.work_p );

  // ʂ߂
  if( ! SysLibFind( sonySysLibNameHR, &hr_ref_num ) )
    {
      HROpen( hr_ref_num );
      HRWinScreenMode( hr_ref_num,
                       winScreenModeSet,
                       &Screen.save.width,
                       &Screen.save.height,
                       &Screen.save.depth,
                       NULL );
      HRClose( hr_ref_num );
    }
  else
    {
      WinScreenMode( winScreenModeSet,
                     &Screen.save.width,
                     &Screen.save.height,
                     &Screen.save.depth,
                     NULL );
    }

  // f[^x[X
  close_geometry_db();
  close_ghost_db();
  close_coursemap_db();

  SysSetAutoOffTime( AutoOff.seconds );
  FrmCloseAllForms();
}



/************************************************
 *	C
 ***********************************************/
static UInt32
my_app_main( UInt16 cmd, MemPtr /*cmdPBP*/, UInt16 launchFlags )
{
  Err err;

  // ROM`FbN OS3.5ȍ~
  err = rom_check( PalmRomVer350, launchFlags );
  if( err ) return err;

  switch( cmd )
    {
    case sysAppLaunchCmdNormalLaunch:
      err = app_start();
      if( err ) return err;

      FrmGotoForm( MainForm );
      event_loop();
      app_stop();
      break;

    default:
      break;
    }

  return errNone;
}



/************************************************
 *	Gg[
 ***********************************************/
UInt32
PilotMain( UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags )
{
  return my_app_main( cmd, cmdPBP, launchFlags );
}



